<?php if (!defined('PmWiki')) exit();
/*
+----------------------------------------------------------------------+
| See cookbook/chordpro/README.txt for information.
| See cookbook/chordpro/LICENSE.txt for licence.
+----------------------------------------------------------------------+
| Copyright 2005 Jeremy H. Sproat
| This program is free software; you can redistribute it and/or modify
| it under the terms of the GNU General Public License, Version 2, as
| published by the Free Software Foundation.
| http://www.gnu.org/copyleft/gpl.html
| This program is distributed in the hope that it will be useful,
| but WITHOUT ANY WARRANTY; without even the implied warranty of
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
| GNU General Public License for more details.
+----------------------------------------------------------------------+
*/


define(CHORDPRO_VERSION, '0.1');


## Uncomment include line to add a custom page storage location 
## and some bundled wikipages.

#@include("chordpro/bundlepages.php");


## Add a custom markup
Markup( 'chordpro', 'fulltext', "/\\|\\|:(.*):\\|\\|/seim",
    "ChordPro_Parse(explode(\"\n\",PSS('$1')))" );
# s = dot matches all chars including newline
# e = backreference substitution by expression
# i = case insensitive
# m = multiline
# TODO: how to disable <pre> blocks?  This needs to be preserved from
# the preformatted block (line begins with space) as well as the
# paragraph (empty line) and the tables (vertical bar).  Is this what 
# <:block> is used for?


## constants for the parser
define("CHORDPRO_BLOCK_NONE", 0);
define("CHORDPRO_BLOCK_TAB", 1);
define("CHORDPRO_BLOCK_LYRICS", 2);
define("CHORDPRO_BLOCK_CHORUS", 3);


/** Main ChordPro parser
 *   /param   $cho  An array of strings containing the ChordPro markup.
 *                  One string per line.  Don't worry about stripping
 *                  whitespace.
 *   /return  The HTML-formatted song wrapped in a <div> of class "song"
 */
function ChordPro_Parse($cho) {

  $retval = '';  
  $blocktype = CHORDPRO_BLOCK_NONE;
  $chordblocks = array();
  $textfont = '';
  $textsize = '';
  $title = '';
  $subtitle = '';
  $chordfont = '';
  $chordsize = '';
  $linecount = count($cho);
  
  if( $linecount <= 1 ) {
    # TODO: debugging message, replace me with something intelligent
    #return '<p><h1>Error: line count = ' . $linecount . '!</h1></p>';
  }
  
  for( $i=0; $i<$linecount; $i++ ) {
    # iterate through the ChordPro lines
    $choline = $cho[$i];
    $matches = array();
    $directive = '';
    $directiveargument = '';
  
    if( preg_match( '/\{([^:]+)(:(.+))?\}/', $choline, $matches ) ) {
      # get directive
      $directive = trim($matches[1]);
      if( count($matches) > 2 ) {
        $directiveargument = trim($matches[3]);
      } else {
        $directiveargument = '';
      }
    }
  
    if( $choline{0} == '#' ) {
      # strip out code comments
      continue;
      
    } else if( $blocktype == CHORDPRO_BLOCK_TAB ) {
      # we're in a tablature block
      # pass through text unprocessed until we see {end_of_tab} or {eot}
      if( strcasecmp($directive,'eot') && 
          strcasecmp($directive,'end_of_tab')
      ) { # i.e., neither token appears...
        if( $choline{0} == ' ' ) {
          # workaround the <pre> block problem
          # TODO: *fix* the problem
          $retval .= '&nbsp;';
          $retval .= substr( $choline, 1, strlen($choline)-1 );
        } else {
          $retval .= $choline;
        }
        $cholineend = $choline{strlen($choline)-1};
        if( $cholineend != "\x0a" && $cholineend != "\x0d" ) {
          # only add EOL if there is none
          $retval .= "\n";
        }
      } else {
        # start processing text
        $retval .= '</div>' . "\n";
        $blocktype = CHORDPRO_BLOCK_NONE;
      }
      
    } else {
      # we're not in a tablature block
        
      if( strlen($directive)>0 ) {
        # process directives
      
        if( !strcasecmp($directive,'sot') || 
            !strcasecmp($directive,'start_of_tab')
        ) {
          # it's a start of a tablature block
          if( $blocktype != CHORDPRO_BLOCK_NONE ) {
            $retval .= '</div>' . "\n";
          }
          $retval .= '<div class="songtab">' . "\n";
          $blocktype = CHORDPRO_BLOCK_TAB;
          
        } else if( !strcasecmp($directive,'define') ) {
          # it's a chord fingering definition line
          # TODO: enable when table formatting is taken care of
          #array_push( $chordblocks, 
          #  ChordPro_ExtractChordBlock($directiveargument) ); 
        
        } else if( !strcasecmp($directive,'rowname') ) {
          # it's a proprietary Songbird Chords directive
          # TODO: enable when table formatting is taken care of
          #array_push( $chordblocks, 
          #  '<br clear="both"/><div class="chordblockrowname songcomment">' . $directiveargument . '</div>' ); 
        
        } else if( !strcasecmp($directive,'textfont') ) {
          # it's a text font name directive
          $textfont = $directiveargument;
        
        } else if( !strcasecmp($directive,'textsize') ) {
          # it's a text font size directive
          $textsize = $directiveargument;
        
        } else if( !strcasecmp($directive,'chordfont') ) {
          # it's a chord notation font name directive
          $chordfont = $directiveargument;
        
        } else if( !strcasecmp($directive,'chordsize') ) {
          # it's a chord notation font size directive
          $chordsize = $directiveargument;
        
        } else if( !strcasecmp($directive,'t') || 
                   !strcasecmp($directive,'title') ) {
          # it's a song title directive
          $title = $directiveargument;

        } else if( !strcasecmp($directive,'st') || 
                   !strcasecmp($directive,'subtitle') ) {
          # it's a song subtitle directive
          $subtitle = $directiveargument;

        } else if( !strcasecmp($directive,'soc') || 
                   !strcasecmp($directive,'start_of_chorus') ) {
          # it's the beginning of a chorus block
          if( $blocktype != CHORDPRO_BLOCK_NONE ) {
            $retval .= '</div>' . "\n";
          }
          $blocktype = CHORDPRO_BLOCK_CHORUS;
          $retval .= '<div class="songchorus">' . "\n";
        
        } else if( !strcasecmp($directive,'eoc') || 
                   !strcasecmp($directive,'end_of_chorus') ) {
          # it's the end of a chorus block
          if( $blocktype != CHORDPRO_BLOCK_NONE ) {
            $retval .= '</div>' . "\n";
          }
          $blocktype = CHORDPRO_BLOCK_LYRICS;
          $retval .= '<div class="songlyric">' . "\n";
        
        } else if( !strcasecmp($directive,'c') || 
                   !strcasecmp($directive,'comment') ) {
          # it's a comment line
          if( $blocktype == CHORDPRO_BLOCK_TAB ) {
            $retval .= '</div>' . "\n";
            $blocktype = CHORDPRO_BLOCK_LYRICS;
            $retval .= '<div class="songlyric">' . "\n";
          }
          if( strlen($directiveargument) > 0 ) {
            $retval .= '<div class="songcomment">' . 
                       $directiveargument . '</div>' . "\n";
          }

        } else if( !strcasecmp($directive,'ci') || 
                   !strcasecmp($directive,'comment_italic') ) {
          # it's an italicized comment line
          if( $blocktype == CHORDPRO_BLOCK_TAB ) {
            $retval .= '</div>' . "\n";
            $blocktype = CHORDPRO_BLOCK_LYRICS;
            $retval .= '<div class="songlyric">' . "\n";
          }
          if( strlen($directiveargument) > 0 ) {
            $retval .= '<div class="songcomment"><i>' . 
                       $directiveargument . '</i></div>' . "\n";
          }

        } else if( !strcasecmp($directive,'cb') || 
                   !strcasecmp($directive,'comment_box') ) {
          # it's a boxed comment line
          if( $blocktype == CHORDPRO_BLOCK_TAB ) {
            $retval .= '</div>' . "\n";
            $blocktype = CHORDPRO_BLOCK_LYRICS;
            $retval .= '<div class="songlyric">' . "\n";
          }
          if( strlen($directiveargument) > 0 ) {
            $retval .= '<div class="songcommentbox">' . 
                       $directiveargument . '</div>' . "\n";
          }

        } else if(
            !strcasecmp($directive,'np') ||
            !strcasecmp($directive,'new_page') ||
            !strcasecmp($directive,'npp') ||
            !strcasecmp($directive,'new_physical_page')
        ) {
          # ignore these directives
          # I may eventually want to follow these directives if the page
          # is put into PDF, though personally I doubt it
          $retval .= '<br/>' . "\n";

        } else if(
            !strcasecmp($directive,'g') ||
            !strcasecmp($directive,'grid') ||
            !strcasecmp($directive,'ns') ||
            !strcasecmp($directive,'new_song') ||
            !strcasecmp($directive,'ng') ||
            !strcasecmp($directive,'no_grid') ||
            !strcasecmp($directive,'tuning')
        ) {
          # ignore these directives
 
        } else {
          # unknown directive
          # TODO: debugging message, replace with something intelligent
          $retval .= '<b color=red>{{' . $directive .
            (strlen($directiveargument)>0?':</b><i>'.$directiveargument.'</i><b>':'') .
            '}}</b>' . '<br/>' . "\n";
          
        }
        
      # I originally wanted to use strpos() to check for the presence of
      # a [ char, intending to format only lyric lines with embedded
      # chords, but PHP's implementation is incredibly bone-headed.
      # If it finds the substring at position 0, it returns FALSE (ok,
      # it's a numerical 0 cast into a boolean FALSE, but come on...)
      # ...but in the end I decided to format all lyric lines this way      
      } else if( strlen(trim($choline)) > 0 ) {
        # format lyrics with embedded chords
        $retval .= ChordPro_ExtractLyricBlocks($choline,$chordfont,$chordsize);
        
      } else {
        # pass through unprocessed unknown text format
        # tacking on a trailing <br/>\n was making PmWiki replace with
        # a <p><br/></p> block in weird places...
        $retval .= $choline;
      }
      
    }
    
  }

  # finished with parsing, let's slap the thing together

  # insert chord blocks
  if( count($chordblocks) > 0 ) {
    $prefix = '<div class="chordblocks">' . "\n";
    foreach( $chordblocks as $block ) {
      $prefix .= $block . "\n";
    }
    $prefix .= '<div style="clear:both;"></div></div>' . "\n";
    $retval = $prefix . $retval;
  }
  
  # insert title block
  if( strlen($title) > 0 ) {
    $prefix = '<div class="songtitle">' . $title;
    if( strlen($subtitle) > 0 ) {
      $prefix .= '<div class="songsubtitle">' . $subtitle . '</div>';
    }
    $prefix .= '</div>' . "\n";
    $retval = $prefix . $retval;
  }
  
  # insert custom font styles
  if ( strlen($textfont) > 0 || strlen($textsize) > 0 ) {
    $prefix = '<div style="' .
      (strlen($textfont)>0?'font-family:\''.$textfont.'\';':'') .
      (strlen($textsize)>0?'font-size:'.$textsize.'px;':'') .
      '">' . "\n";
    $suffix = '</div>' . "\n";
    $retval = $prefix . $retval . $suffix;
  }
  
  return '<div class="song">' . "\n" . $retval . '</div>';
}


/** ChordPro parser helper function - parse chord definition blocks
 *   /param   $defline  A string: the definition line minus the
 *                      "define:" keyword and surrounding braces
 *   /return  The HTML-formatted chord in a div of class "chordblock"
 */
function ChordPro_ExtractChordBlock($defline) {
  $retval = '';
  $chordblock = '';
  $matches = array();
  $stringarray = array();
  $chordname = '';
  $basefret = '';
  $highfret = 0;
  $stringset = '';
  $fret = '';
  $finger = '';
  $stringrows = array(
    '======',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||',
    '||||||'
  );
  
  if( preg_match( '/(\S+) +base-fret +([0-9]+) +frets? +(.+)/i',
      $defline, $matches ) ) {
    # looks like a valid define: directive
    $chordname = trim($matches[1]);
    $basefret = trim($matches[2]);
    $stringset = trim($matches[3]);
    if( $basefret != 0 ) {
      # top of graph is not fret 0, change notation
      $stringrows[0] = '------';
    }
    $highfret = 1;
    
    # split the guitar strings into an array
    $stringarray = preg_split( '/[ _=]+/i', $stringset );
    if( count($stringarray) > 1 ) {
      for( $i = 0; $i < count($stringarray); $i++ ) {
        $str = $stringarray[$i];
        
        if( !strcmp( $str, '-' ) ) {
          # no info for this string, ignore
          #           
        } else if( !strcasecmp( $str, 'x' ) ) {
          # muted string
          $str = 'x';
          $stringrows[0]{$i} = $str;
          
        } else if( !strcmp( $str, '0' ) || !strcasecmp( $str, 'o' ) ) {
          # open string
          $str = 'o'; # TODO: make this configurable          
          $stringrows[0]{$i} = $str;
          
        } else {
          # fret position and finger
          if( preg_match( '/([x0-9]+)([imrpt])?/i', $str, $matches ) ) {
            $fret = $matches[1];
            if( count($matches) > 2 ) {
              $finger = $matches[2];
            }
            $finger = 'o'; # TODO: make this configurable
            if( !strcasecmp($fret,'x') ) {
              $fret = 0;
              $finger = 'x';
            }
            # put the fingering notation into the graph
            $stringrows[$fret]{$i} = $finger;
            if( $fret > $highfret ) {
              $highfret = $fret;
            }
          } else {
            # TODO: debugging message, replace with something intelligent
            return '<div class="chordblock">Malformed str: ' .
              $str . ' within ' . $defline . '</div>';
          }
        }
        
      }
      
    } else {
      # TODO: debugging message, replace with something intelligent
      return '<div class="chordblock">Malformed stringset: ' .
        $stringset . ' within ' . $defline . '</div>';
    }
    
    # tack on base-fret if not at fret 0
    if( $basefret != 0 ) {
      $stringrows[1] .= ' ' . $basefret . 'fr.';
    }
    
    # merge rows into one string
    for( $i = 0; $i <= $highfret + 1; $i++ ) {
      $chordblock .= $stringrows[$i] . "\n";
    }
    
    # center chord name if not too long
    if( strlen($chordname) < 6 ) {
      $chordname = str_pad( $chordname, 6, ' ', STR_PAD_BOTH );
    }
    
  } else {
    # TODO: debugging message, replace with something intelligent
    return '<div class="chordblock">Malformed define: ' . $defline . '</div>';
  }
  
  # format for html
  $retval = '<div class="chordblock">' . $chordname . "\n" . $chordblock . '</div>';
  return $retval;
}


/** ChordPro parser helper function - extract and parse lyrics with embedded chord notation
 *   /param   $choline  A string: the chordpro lyric line with embedded chords
 *   /param   $chordfont  A string: the optional chord font.  Set to '' for default.
 *   /param   $chordsize  A string: the optional chord font size.  Set to '' to disable.
 *   /return  The HTML-formatted lyric line
 */
function ChordPro_ExtractLyricBlocks($choline,$chordfont,$chordsize) {
  # Pull the chord notation from a line of lyrics and return a formatted table
  # ...each column is a segment of lyrics with its accompanying chord above it
  # ...in general, a new column is created when a chord appears

  # You [Em]stole my heart and [A]that's what [G]really [D]hurts [A] [B] [C]
  #   ...becomes...
  # |    |[Em]               |[A]         |[G]    |[D]   |[A] |[B] |[C] |
  # |You |stole my heart and |that's what |really |hurts |    |    |    |

  # Ugly css hack with chordfont and chordsize in here somewhere
  
  $choline = trim($choline);
  $inchord = false;
  $chords = array();
  $lyrics = array();
  $col = 0;
  array_push( $chords, '' );
  array_push( $lyrics, '' );
  for( $i = 0; $i < strlen($choline); $i++ ) {
    $ch = $choline{$i};
    if($ch=='[') {
      # at start of chord boundry
      if($inchord) {
        # this no-op intentionally left blank
      } else {
        $inchord = true;
        $col++;
        array_push( $chords, '' );
        array_push( $lyrics, '' );
      }
    } else if($ch==']') {
      # at end of chord boundry
      if($inchord) {
        $chords[$col] .= ' ';
        $inchord = false;
      } else {
        # this no-op intentionally left blank
      }
    } else {
      # not at boundry
      if($inchord) {
        $chords[$col] .= $ch;
      } else {
        $lyrics[$col] .= $ch;
      }      
    }
  }
  $retval = '<table border=0 cellpadding=0 cellspacing=0>';
  $retval .= '<tr class="songlyricchord">';
  for( $i = 0; $i <= $col; $i++ ) {
    $retval .= '<td';
    if( strlen($chordfont)>0 || strlen($chordsize)>0 ) {
      $retval .= ' style="';
      if( strlen($chordfont)>0 ) {
        $retval .= 'font-family:\'' . $chordfont . '\';';
      }
      if( strlen($chordsize)>0 ) {
        $retval .= 'font-size:' . $chordsize . ';';
      }
      $retval .= '"';
    }
    $retval .= '>' . $chords[$i] . '</td>';
  }
  $retval .= '</tr><tr class="songlyricline">';
  for( $i = 0; $i < $col+1; $i++ ) {
    $retval .= '<td>' . $lyrics[$i] . '</td>';
  }
  $retval .= '</table>' . "\n";
  return $retval;
}


/** ChordPro parser test function - read a file and parse it
 *   /param   $chofile  name of ChordPro file to read
 *   /return  The file as sent through the parser
 */
function ChordPro_ParseFile($chofile) {
  $handle = fopen($chofile, 'r');
  $cho = array();
  if($handle) {
    while(!feof($handle)) {
      array_push( $cho, fgets($handle) );
    }
    fclose($handle);
    return ChordPro_Parse($cho);
  } else {
    fclose($chofile);
    return '<p><h1>Error opening file!</h1></p>';
    return -1;
  }
}
